@@ -15,13 +15,17 @@
 //
 // Copyright (C) 2006 Scott Turner <scotty1024@mac.com>
 // Copyright (C) 2007, 2008 Julien Rebetez <julienr@svn.gnome.org>
-// Copyright (C) 2007-2010 Carlos Garcia Campos <carlosgc@gnome.org>
+// Copyright (C) 2007-2011, 2013, 2015 Carlos Garcia Campos <carlosgc@gnome.org>
 // Copyright (C) 2007, 2008 Iñigo Martínez <inigomartinez@gmail.com>
 // Copyright (C) 2008 Michael Vrable <mvrable@cs.ucsd.edu>
 // Copyright (C) 2008 Hugo Mercier <hmercier31@gmail.com>
 // Copyright (C) 2008 Pino Toscano <pino@kde.org>
 // Copyright (C) 2008 Tomas Are Haavet <tomasare@gmail.com>
-// Copyright (C) 2009 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2009-2011, 2013, 2016 Albert Astals Cid <aacid@kde.org>
+// Copyright (C) 2012, 2013 Fabio D'Urso <fabiodurso@hotmail.it>
+// Copyright (C) 2012, 2015 Tobias Koenig <tokoe@kdab.com>
+// Copyright (C) 2013 Thomas Freitag <Thomas.Freitag@alfa.de>
+// Copyright (C) 2013 Adrian Johnson <ajohnson@redneon.com>
 //
 // To see a description of the changes please see the Changelog file that
 // came with your tarball or type make ChangeLog if you are building from git
@@ -35,19 +39,24 @@
 #pragma interface
 #endif
 
+#include "Object.h"
+
 class XRef;
 class Gfx;
-class Catalog;
 class CharCodeToUnicode;
 class GfxFont;
-class GfxFontDict;
+class GfxResources;
+class Page;
+class PDFDoc;
 class Form;
 class FormWidget;
+class FormField;
+class FormFieldChoice;
 class PDFRectangle;
 class Movie;
 class LinkAction;
-class OCGs;
 class Sound;
+class FileSpec;
 
 enum AnnotLineEndingStyle {
   annotLineEndingSquare,        // Square
@@ -173,6 +182,7 @@ private:
 //------------------------------------------------------------------------
 
 class AnnotQuadrilaterals {
+public:
   class AnnotQuadrilateral {
   public:
     AnnotQuadrilateral(double x1, double y1, double x2, double y2, double x3,
@@ -181,9 +191,8 @@ class AnnotQuadrilaterals {
     AnnotCoord coord1, coord2, coord3, coord4;
   };
 
-public:
-
   AnnotQuadrilaterals(Array *array, PDFRectangle *rect);
+  AnnotQuadrilaterals(AnnotQuadrilateral **quads, int quadsLength);
   ~AnnotQuadrilaterals();
 
   double getX1(int quadrilateral);
@@ -208,7 +217,6 @@ protected:
 class AnnotBorder {
 public:
   enum AnnotBorderType {
-    typeUnknown,
     typeArray,
     typeBS
   };
@@ -221,18 +229,26 @@ public:
     borderUnderlined  // Underlined
   };
 
-  AnnotBorder();
   virtual ~AnnotBorder();
 
-  virtual AnnotBorderType getType() const { return type; }
+  virtual void setWidth(double new_width) { width = new_width; }
+
+  virtual AnnotBorderType getType() const = 0;
   virtual double getWidth() const { return width; }
   virtual int getDashLength() const { return dashLength; }
   virtual double *getDash() const { return dash; }
   virtual AnnotBorderStyle getStyle() const { return style; }
 
+  virtual void writeToObject(XRef *xref, Object *obj1) const = 0;
+
 protected:
+  AnnotBorder();
+
+  GBool parseDashArray(Object *dashObj);
+
   AnnotBorderType type;
   double width;
+  static const int DASH_LIMIT = 10; // implementation note 82 in Appendix H.
   int dashLength;
   double *dash;
   AnnotBorderStyle style;
@@ -247,11 +263,16 @@ public:
   AnnotBorderArray();
   AnnotBorderArray(Array *array);
 
+  void setHorizontalCorner(double hc) { horizontalCorner = hc; }
+  void setVerticalCorner(double vc) { verticalCorner = vc; }
+
   double getHorizontalCorner() const { return horizontalCorner; }
   double getVerticalCorner() const { return verticalCorner; }
 
-protected:
-  static const int DASH_LIMIT = 10; // implementation note 82 in Appendix H.
+private:
+  virtual AnnotBorderType getType() const { return typeArray; }
+  virtual void writeToObject(XRef *xref, Object *obj1) const;
+
   double horizontalCorner;          // (Default 0)
   double verticalCorner;            // (Default 0)
   // double width;                  // (Default 1)  (inherited from AnnotBorder)
@@ -268,6 +289,11 @@ public:
   AnnotBorderBS(Dict *dict);
 
 private:
+  virtual AnnotBorderType getType() const { return typeBS; }
+  virtual void writeToObject(XRef *xref, Object *obj1) const;
+
+  const char *getStyleName() const;
+
   // double width;           // W  (Default 1)   (inherited from AnnotBorder)
   // AnnotBorderStyle style; // S  (Default S)   (inherited from AnnotBorder)
   // double *dash;           // D  (Default [3]) (inherited from AnnotBorder)
@@ -293,51 +319,19 @@ public:
   AnnotColor(double c, double m, double y, double k);
   AnnotColor(Array *array, int adjust = 0);
 
+  void adjustColor(int adjust);
+
   AnnotColorSpace getSpace() const { return (AnnotColorSpace) length; }
   const double *getValues() const { return values; }
 
+  void writeToObject(XRef *xref, Object *dest) const;
+
 private:
 
   double values[4];
   int length;
 };
 
-//------------------------------------------------------------------------
-// AnnotBorderStyle
-//------------------------------------------------------------------------
-
-enum AnnotBorderType {
-  annotBorderSolid,
-  annotBorderDashed,
-  annotBorderBeveled,
-  annotBorderInset,
-  annotBorderUnderlined
-};
-
-class AnnotBorderStyle {
-public:
-
-  AnnotBorderStyle(AnnotBorderType typeA, double widthA,
-		   double *dashA, int dashLengthA,
-		   double rA, double gA, double bA);
-  ~AnnotBorderStyle();
-
-  AnnotBorderType getType() { return type; }
-  double getWidth() { return width; }
-  void getDash(double **dashA, int *dashLengthA)
-    { *dashA = dash; *dashLengthA = dashLength; }
-  void getColor(double *rA, double *gA, double *bA)
-    { *rA = r; *gA = g; *bA = b; }
-
-private:
-
-  AnnotBorderType type;
-  double width;
-  double *dash;
-  int dashLength;
-  double r, g, b;
-};
-
 //------------------------------------------------------------------------
 // AnnotIconFit
 //------------------------------------------------------------------------
@@ -387,7 +381,32 @@ public:
     appearDown
   };
 
-  AnnotAppearance(Dict *dict);
+  AnnotAppearance(PDFDoc *docA, Object *dict);
+  ~AnnotAppearance();
+
+  // State is ignored if no subdictionary is present
+  void getAppearanceStream(AnnotAppearanceType type, const char *state, Object *dest);
+
+  // Access keys in normal appearance subdictionary (N)
+  GooString * getStateKey(int i);
+  int getNumStates();
+
+  // Removes all associated streams in the xref table. Caller is required to
+  // reset parent annotation's AP and AS after this call.
+  void removeAllStreams();
+
+  // Test if this AnnotAppearance references the specified stream
+  GBool referencesStream(Ref targetStreamRef);
+
+private:
+  static GBool referencesStream(Object *stateObj, Ref targetStreamRef);
+  void removeStream(Ref refToStream);
+  void removeStateStreams(Object *state);
+
+protected:
+  PDFDoc *doc;
+  XRef *xref;                   // the xref table for this PDF file
+  Object appearDict;            // Annotation's AP
 };
 
 //------------------------------------------------------------------------
@@ -434,11 +453,39 @@ protected:
   AnnotAppearanceCharacsTextPos position; // TP (Default 0)
 };
 
+//------------------------------------------------------------------------
+// AnnotAppearanceBBox
+//------------------------------------------------------------------------
+
+class AnnotAppearanceBBox
+{
+public:
+  AnnotAppearanceBBox(PDFRectangle *init);
+
+  void setBorderWidth(double w) { borderWidth = w; }
+
+  // The following functions operate on coords relative to [origX origY]
+  void extendTo(double x, double y);
+  void getBBoxRect(double bbox[4]) const;
+
+  // Get boundaries in page coordinates
+  double getPageXMin() const;
+  double getPageYMin() const;
+  double getPageXMax() const;
+  double getPageYMax() const;
+
+private:
+  double origX, origY, borderWidth;
+  double minX, minY, maxX, maxY;
+};
+
 //------------------------------------------------------------------------
 // Annot
 //------------------------------------------------------------------------
 
 class Annot {
+  friend class Annots;
+  friend class Page;
 public:
   enum AnnotFlag {
     flagUnknown        = 0x0000,
@@ -480,45 +527,87 @@ public:
     typePrinterMark,    // PrinterMark    22
     typeTrapNet,        // TrapNet        23
     typeWatermark,      // Watermark      24
-    type3D              // 3D             25
+    type3D,             // 3D             25
+    typeRichMedia       // RichMedia      26
   };
 
-  Annot(XRef *xrefA, PDFRectangle *rectA, Catalog *catalog);
-  Annot(XRef *xrefA, Dict *dict, Catalog *catalog);
-  Annot(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
-  virtual ~Annot();
+  /**
+   * Describes the additional actions of a screen or widget annotation.
+   */
+  enum AdditionalActionsType {
+    actionCursorEntering, ///< Performed when the cursor enters the annotation's active area
+    actionCursorLeaving,  ///< Performed when the cursor exists the annotation's active area
+    actionMousePressed,   ///< Performed when the mouse button is pressed inside the annotation's active area
+    actionMouseReleased,  ///< Performed when the mouse button is released inside the annotation's active area
+    actionFocusIn,        ///< Performed when the annotation receives the input focus
+    actionFocusOut,       ///< Performed when the annotation loses the input focus
+    actionPageOpening,    ///< Performed when the page containing the annotation is opened
+    actionPageClosing,    ///< Performed when the page containing the annotation is closed
+    actionPageVisible,    ///< Performed when the page containing the annotation becomes visible
+    actionPageInvisible   ///< Performed when the page containing the annotation becomes invisible
+  };
+
+  enum FormAdditionalActionsType {
+    actionFieldModified,   ///< Performed when the when the user modifies the field
+    actionFormatField,     ///< Performed before the field is formatted to display its value
+    actionValidateField,   ///< Performed when the field value changes
+    actionCalculateField,  ///< Performed when the field needs to be recalculated
+  };
+
+  Annot(PDFDoc *docA, PDFRectangle *rectA);
+  Annot(PDFDoc *docA, Dict *dict);
+  Annot(PDFDoc *docA, Dict *dict, Object *obj);
   GBool isOk() { return ok; }
 
+  void incRefCnt();
+  void decRefCnt();
+
   virtual void draw(Gfx *gfx, GBool printing);
-  // Get appearance object.
-  Object *getAppearance(Object *obj) { return appearance.fetch(xref, obj); }
+  // Get the resource dict of the appearance stream
+  virtual Object *getAppearanceResDict(Object *dest);
 
   GBool match(Ref *refA)
     { return ref.num == refA->num && ref.gen == refA->gen; }
 
   double getXMin();
   double getYMin();
+  double getXMax();
+  double getYMax();
 
   double getFontSize() { return fontSize; }
 
+  void setRect(PDFRectangle *rect);
+  void setRect(double x1, double y1, double x2, double y2);
+
   // Sets the annot contents to new_content
   // new_content should never be NULL
-  void setContents(GooString *new_content);
+  virtual void setContents(GooString *new_content);
+  void setName(GooString *new_name);
+  void setModified(GooString *new_date);
+  void setFlags(Guint new_flags);
+
+  void setBorder(AnnotBorder *new_border); // Takes ownership
 
   // The annotation takes the ownership of
   // new_color. 
   void setColor(AnnotColor *new_color);
 
+  void setAppearanceState(const char *state);
+
   // getters
+  PDFDoc *getDoc() const { return doc; }
+  XRef *getXRef() const { return xref; }
+  GBool getHasRef() const { return hasRef; }
   Ref getRef() const { return ref; }
   AnnotSubtype getType() const { return type; }
   PDFRectangle *getRect() const { return rect; }
+  void getRect(double *x1, double *y1, double *x2, double *y2) const;
   GooString *getContents() const { return contents; }
   int getPageNum() const { return page; }
   GooString *getName() const { return name; }
   GooString *getModified() const { return modified; }
   Guint getFlags() const { return flags; }
-  /*Dict *getAppearDict() const { return appearDict; }*/
+  AnnotAppearance *getAppearStreams() const { return appearStreams; }
   GooString *getAppearState() const { return appearState; }
   AnnotBorder *getBorder() const { return border; }
   AnnotColor *getColor() const { return color; }
@@ -526,27 +615,44 @@ public:
 
   int getId() { return ref.num; }
 
+  // Check if point is inside the annot rectangle.
+  GBool inRect(double x, double y) const;
+
 private:
   void readArrayNum(Object *pdfArray, int key, double *value);
   // write vStr[i:j[ in appearBuf
 
-  void initialize (XRef *xrefA, Dict *dict, Catalog *catalog);
+  void initialize (PDFDoc *docA, Dict *dict);
+  void setPage (int new_page, GBool updateP); // Called by Page::addAnnot and Annots ctor
 
 
 protected:
+  virtual ~Annot();
+  virtual void removeReferencedObjects(); // Called by Page::removeAnnot
   void setColor(AnnotColor *color, GBool fill);
+  void setLineStyleForBorder(AnnotBorder *border);
   void drawCircle(double cx, double cy, double r, GBool fill);
   void drawCircleTopLeft(double cx, double cy, double r);
   void drawCircleBottomRight(double cx, double cy, double r);
+  void layoutText(GooString *text, GooString *outBuf, int *i, GfxFont *font,
+		  double *width, double widthLimit, int *charCount,
+		  GBool noReencode);
+  void writeString(GooString *str, GooString *appearBuf);
   void createForm(double *bbox, GBool transparencyGroup, Object *resDict, Object *aStream);
-  void createResourcesDict(char *formName, Object *formStream, char *stateName,
-			   double opacity, char *blendMode, Object *resDict);
+  void createResourcesDict(const char *formName, Object *formStream, const char *stateName,
+			   double opacity, const char *blendMode, Object *resDict);
   GBool isVisible(GBool printing);
+  int getRotation() const;
 
   // Updates the field key of the annotation dictionary
   // and sets M to the current time
   void update(const char *key, Object *value);
 
+  // Delete appearance streams and reset appearance state
+  void invalidateAppearance();
+
+  int refCnt;
+
   Object annotObj;
   
   // required data
@@ -559,15 +665,15 @@ protected:
   GooString *name;                  // NM
   GooString *modified;              // M
   Guint flags;                      // F (must be a 32 bit unsigned int)
-  //Dict *appearDict;                 // AP (should be correctly parsed)
-  Ref appRef;                       //the reference to the indirect appearance object in XRef 
+  AnnotAppearance *appearStreams;   // AP
   Object appearance;     // a reference to the Form XObject stream
                          //   for the normal appearance
+  AnnotAppearanceBBox *appearBBox;  // BBox of generated appearance
   GooString *appearState;           // AS
   int treeKey;                      // Struct Parent;
-  OCGs *optContentConfig;           // Optional content config
   Object oc;                        // OC
 
+  PDFDoc *doc;
   XRef *xref;			// the xref table for this PDF file
   Ref ref;                      // object ref identifying this annotation
   GooString *appearBuf;
@@ -577,6 +683,9 @@ protected:
   GBool ok;
 
   bool hasRef;
+#if MULTITHREADED
+  GooMutex mutex;
+#endif
 };
 
 //------------------------------------------------------------------------
@@ -585,19 +694,19 @@ protected:
 
 class AnnotPopup: public Annot {
 public:
-  AnnotPopup(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  AnnotPopup(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotPopup(PDFDoc *docA, PDFRectangle *rect);
+  AnnotPopup(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotPopup();
 
   Object *getParent(Object *obj) { return parent.fetch (xref, obj); }
-  Object *getParentNF(Object *obj) { return &parent; }
+  Object *getParentNF() { return &parent; }
   void setParent(Object *parentA);
   void setParent(Annot *parentA);
   GBool getOpen() const { return open; }
   void setOpen(GBool openA);
 
 protected:
-  void initialize(XRef *xrefA, Dict *dict, Catalog *catalog);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   Object parent; // Parent
   GBool open;   // Open
@@ -614,8 +723,8 @@ public:
     replyTypeGroup  // Group
   };
 
-  AnnotMarkup(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  AnnotMarkup(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotMarkup(PDFDoc *docA, PDFRectangle *rect);
+  AnnotMarkup(PDFDoc *docA, Dict *dict, Object *obj);
   virtual ~AnnotMarkup();
 
   // getters
@@ -632,8 +741,12 @@ public:
   // The annotation takes the ownership of new_popup
   void setPopup(AnnotPopup *new_popup);
   void setLabel(GooString *new_label);
+  void setOpacity(double opacityA);
+  void setDate(GooString *new_date);
 
 protected:
+  virtual void removeReferencedObjects();
+
   GooString *label;             // T            (Default autor)
   AnnotPopup *popup;            // Popup
   double opacity;               // CA           (Default 1.0)
@@ -648,7 +761,7 @@ protected:
   AnnotExternalDataType exData; // ExData
 
 private:
-  void initialize(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  void initialize(PDFDoc *docA, Dict *dict, Object *obj);
 };
 
 //------------------------------------------------------------------------
@@ -670,8 +783,8 @@ public:
     stateNone       // None
   };
 
-  AnnotText(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  AnnotText(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotText(PDFDoc *docA, PDFRectangle *rect);
+  AnnotText(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotText();
 
   virtual void draw(Gfx *gfx, GBool printing);
@@ -686,7 +799,7 @@ public:
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   GBool open;                       // Open       (Default false)
   GooString *icon;                  // Name       (Default Note)
@@ -703,8 +816,8 @@ private:
 
 class AnnotMovie: public Annot {
  public:
-  AnnotMovie(XRef *xrefA, PDFRectangle *rect, Movie *movieA, Catalog *catalog);
-  AnnotMovie(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotMovie(PDFDoc *docA, PDFRectangle *rect, Movie *movieA);
+  AnnotMovie(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotMovie();
 
   virtual void draw(Gfx *gfx, GBool printing);
@@ -713,7 +826,7 @@ class AnnotMovie: public Annot {
   Movie* getMovie() { return movie; }
 
  private:
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   GooString* title;      // T
   Movie* movie;          // Movie + A
@@ -727,18 +840,18 @@ class AnnotMovie: public Annot {
 class AnnotScreen: public Annot {
  public:
 
-  AnnotScreen(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  AnnotScreen(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotScreen(PDFDoc *docA, PDFRectangle *rect);
+  AnnotScreen(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotScreen();
 
   GooString* getTitle() { return title; }
 
   AnnotAppearanceCharacs *getAppearCharacs() { return appearCharacs; }
-  LinkAction* getAction() { return action; }
-  Object* getAdditionActions() { return &additionAction; }
+  LinkAction* getAction() { return action; } // The caller should now delete the result
+  LinkAction *getAdditionalAction(AdditionalActionsType type); // The caller should delete the result
 
  private:
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
 
   GooString* title;                      // T
@@ -746,7 +859,7 @@ class AnnotScreen: public Annot {
   AnnotAppearanceCharacs* appearCharacs; // MK
 
   LinkAction *action;                    // A
-  Object additionAction;                 // AA
+  Object additionalActions;              // AA
 };
 
 //------------------------------------------------------------------------
@@ -763,25 +876,23 @@ public:
     effectPush      // P
   };
 
-  AnnotLink(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  AnnotLink(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotLink(PDFDoc *docA, PDFRectangle *rect);
+  AnnotLink(PDFDoc *docA, Dict *dict, Object *obj);
   virtual ~AnnotLink();
 
   virtual void draw(Gfx *gfx, GBool printing);
 
   // getters
-  Dict *getActionDict() const { return actionDict; }
-  Object *getDest() { return &dest; }
+  LinkAction *getAction() const { return action; }
   AnnotLinkEffect getLinkEffect() const { return linkEffect; }
   Dict *getUriAction() const { return uriAction; }
   AnnotQuadrilaterals *getQuadrilaterals() const { return quadrilaterals; }
 
 protected:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
-  Dict *actionDict;                    // A
-  Object dest;                         // Dest
+  LinkAction *action;                  // A, Dest
   AnnotLinkEffect linkEffect;          // H          (Default I)
   Dict *uriAction;                     // PA
 
@@ -807,10 +918,20 @@ public:
     intentFreeTextTypeWriter  // FreeTextTypeWriter
   };
 
-  AnnotFreeText(XRef *xrefA, PDFRectangle *rect, GooString *da, Catalog *catalog);
-  AnnotFreeText(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotFreeText(PDFDoc *docA, PDFRectangle *rect, GooString *da);
+  AnnotFreeText(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotFreeText();
 
+  virtual void draw(Gfx *gfx, GBool printing);
+  virtual Object *getAppearanceResDict(Object *dest);
+  virtual void setContents(GooString *new_content);
+
+  void setAppearanceString(GooString *new_string);
+  void setQuadding(AnnotFreeTextQuadding new_quadding);
+  void setStyleString(GooString *new_string);
+  void setCalloutLine(AnnotCalloutLine *line);
+  void setIntent(AnnotFreeTextIntent new_intent);
+
   // getters
   GooString *getAppearanceString() const { return appearanceString; }
   AnnotFreeTextQuadding getQuadding() const { return quadding; }
@@ -824,7 +945,9 @@ public:
 
 protected:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
+  static void parseAppearanceString(GooString *da, double &fontsize, AnnotColor* &fontcolor);
+  void generateFreeTextAppearance();
 
   // required
   GooString *appearanceString;      // DA
@@ -859,11 +982,21 @@ public:
     captionPosTop     // Top
   };
 
-  AnnotLine(XRef *xrefA, PDFRectangle *rect, PDFRectangle *lRect, Catalog *catalog);
-  AnnotLine(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotLine(PDFDoc *docA, PDFRectangle *rect);
+  AnnotLine(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotLine();
 
   virtual void draw(Gfx *gfx, GBool printing);
+  virtual Object *getAppearanceResDict(Object *dest);
+  virtual void setContents(GooString *new_content);
+
+  void setVertices(double x1, double y1, double x2, double y2);
+  void setStartEndStyle(AnnotLineEndingStyle start, AnnotLineEndingStyle end);
+  void setInteriorColor(AnnotColor *new_color);
+  void setLeaderLineLength(double len);
+  void setLeaderLineExtension(double len);
+  void setCaption(bool new_cap);
+  void setIntent(AnnotLineIntent new_intent);
 
   // getters
   AnnotLineEndingStyle getStartStyle() const { return startStyle; }
@@ -885,7 +1018,8 @@ public:
 
 protected:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
+  void generateLineAppearance();
 
   // required
   AnnotCoord *coord1, *coord2;
@@ -914,18 +1048,22 @@ protected:
 class AnnotTextMarkup: public AnnotMarkup {
 public:
 
-  AnnotTextMarkup(XRef *xrefA, PDFRectangle *rect, AnnotSubtype subType,
-		  AnnotQuadrilaterals *quadPoints, Catalog *catalog);
-  AnnotTextMarkup(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotTextMarkup(PDFDoc *docA, PDFRectangle *rect, AnnotSubtype subType);
+  AnnotTextMarkup(PDFDoc *docA, Dict *dict, Object *obj);
   virtual ~AnnotTextMarkup();
 
   virtual void draw(Gfx *gfx, GBool printing);
 
+  // typeHighlight, typeUnderline, typeSquiggly or typeStrikeOut
+  void setType(AnnotSubtype new_type);
+
+  void setQuadrilaterals(AnnotQuadrilaterals *quadPoints);
+
   AnnotQuadrilaterals *getQuadrilaterals() const { return quadrilaterals; }
 
 protected:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
   
   AnnotQuadrilaterals *quadrilaterals; // QuadPoints
 };
@@ -937,16 +1075,18 @@ protected:
 class AnnotStamp: public AnnotMarkup {
 public:
 
-  AnnotStamp(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  AnnotStamp(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotStamp(PDFDoc *docA, PDFRectangle *rect);
+  AnnotStamp(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotStamp();
 
+  void setIcon(GooString *new_icon);
+
   // getters
   GooString *getIcon() const { return icon; }
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   GooString *icon;                  // Name       (Default Draft)
 };
@@ -958,12 +1098,15 @@ private:
 class AnnotGeometry: public AnnotMarkup {
 public:
 
-  AnnotGeometry(XRef *xrefA, PDFRectangle *rect, AnnotSubtype subType, Catalog *catalog);
-  AnnotGeometry(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotGeometry(PDFDoc *docA, PDFRectangle *rect, AnnotSubtype subType);
+  AnnotGeometry(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotGeometry();
 
   virtual void draw(Gfx *gfx, GBool printing);
 
+  void setType(AnnotSubtype new_type); // typeSquare or typeCircle
+  void setInteriorColor(AnnotColor *new_color);
+
   // getters
   AnnotColor *getInteriorColor() const { return interiorColor; }
   AnnotBorderEffect *getBorderEffect() const { return borderEffect; }
@@ -971,7 +1114,7 @@ public:
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   AnnotColor *interiorColor;        // IC
   AnnotBorderEffect *borderEffect;  // BE
@@ -991,11 +1134,18 @@ public:
     polygonDimension   // PolygonDimension
   };
 
-  AnnotPolygon(XRef *xrefA, PDFRectangle *rect, AnnotSubtype subType,
-	       AnnotPath *path, Catalog *catalog);
-  AnnotPolygon(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotPolygon(PDFDoc *docA, PDFRectangle *rect, AnnotSubtype subType);
+  AnnotPolygon(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotPolygon();
 
+  virtual void draw(Gfx *gfx, GBool printing);
+
+  void setType(AnnotSubtype new_type); // typePolygon or typePolyLine
+  void setVertices(AnnotPath *path);
+  void setStartEndStyle(AnnotLineEndingStyle start, AnnotLineEndingStyle end);
+  void setInteriorColor(AnnotColor *new_color);
+  void setIntent(AnnotPolygonIntent new_intent);
+
   // getters
   AnnotPath *getVertices() const { return vertices; }
   AnnotLineEndingStyle getStartStyle() const { return startStyle; }
@@ -1006,7 +1156,7 @@ public:
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   // required
   AnnotPath *vertices;              // Vertices
@@ -1034,17 +1184,19 @@ public:
     symbolP         // P
   };
 
-  AnnotCaret(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  AnnotCaret(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotCaret(PDFDoc *docA, PDFRectangle *rect);
+  AnnotCaret(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotCaret();
 
+  void setSymbol(AnnotCaretSymbol new_symbol);
+
   // getters
   AnnotCaretSymbol getSymbol() const { return symbol; }
   PDFRectangle *getCaretRect() const { return caretRect; }
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   AnnotCaretSymbol symbol;       // Sy         (Default None)
   PDFRectangle *caretRect;       // RD (combined with Rect)
@@ -1057,17 +1209,24 @@ private:
 class AnnotInk: public AnnotMarkup {
 public:
 
-  AnnotInk(XRef *xrefA, PDFRectangle *rect, AnnotPath **paths, int n_paths, Catalog *catalog);
-  AnnotInk(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotInk(PDFDoc *docA, PDFRectangle *rect);
+  AnnotInk(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotInk();
 
+  virtual void draw(Gfx *gfx, GBool printing);
+
+  void setInkList(AnnotPath **paths, int n_paths);
+
   // getters
   AnnotPath **getInkList() const { return inkList; }
   int getInkListLength() const { return inkListLength; }
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
+  void writeInkList(AnnotPath **paths, int n_paths, Array *dest_array);
+  void parseInkList(Array *src_array);
+  void freeInkList();
 
   // required
   AnnotPath **inkList;       // InkList
@@ -1085,8 +1244,8 @@ private:
 class AnnotFileAttachment: public AnnotMarkup {
 public:
 
-  AnnotFileAttachment(XRef *xrefA, PDFRectangle *rect, GooString *filename, Catalog *catalog);
-  AnnotFileAttachment(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotFileAttachment(PDFDoc *docA, PDFRectangle *rect, GooString *filename);
+  AnnotFileAttachment(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotFileAttachment();
 
   virtual void draw(Gfx *gfx, GBool printing);
@@ -1097,7 +1256,7 @@ public:
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   // required
   Object file;      // FS
@@ -1113,8 +1272,8 @@ private:
 class AnnotSound: public AnnotMarkup {
 public:
 
-  AnnotSound(XRef *xrefA, PDFRectangle *rect, Sound *soundA, Catalog *catalog);
-  AnnotSound(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotSound(PDFDoc *docA, PDFRectangle *rect, Sound *soundA);
+  AnnotSound(PDFDoc *docA, Dict *dict, Object *obj);
   ~AnnotSound();
 
   virtual void draw(Gfx *gfx, GBool printing);
@@ -1125,7 +1284,7 @@ public:
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   // required
   Sound *sound;                  // Sound
@@ -1148,46 +1307,48 @@ public:
     highlightModePush     // P,T
   };
 
-  AnnotWidget(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  AnnotWidget(PDFDoc *docA, Dict *dict, Object *obj);
+  AnnotWidget(PDFDoc *docA, Dict *dict, Object *obj, FormField *fieldA);
   virtual ~AnnotWidget();
 
   virtual void draw(Gfx *gfx, GBool printing);
 
+  void drawBorder();
+  void drawFormFieldButton(GfxResources *resources, GooString *da);
+  void drawFormFieldText(GfxResources *resources, GooString *da);
+  void drawFormFieldChoice(GfxResources *resources, GooString *da);
   void generateFieldAppearance ();
+  void updateAppearanceStream ();
 
   AnnotWidgetHighlightMode getMode() { return mode; }
   AnnotAppearanceCharacs *getAppearCharacs() { return appearCharacs; }
-  Dict *getAction() { return action; }
-  Dict *getAdditionActions() { return additionActions; }
+  LinkAction *getAction() { return action; }  // The caller should not delete the result
+  LinkAction *getAdditionalAction(AdditionalActionsType type); // The caller should delete the result
+  LinkAction *getFormAdditionalAction(FormAdditionalActionsType type); // The caller should delete the result
   Dict *getParent() { return parent; }
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
-  void drawText(GooString *text, GooString *da, GfxFontDict *fontDict,
+  void drawText(GooString *text, GooString *da, GfxResources *resources,
 		GBool multiline, int comb, int quadding,
 		GBool txField, GBool forceZapfDingbats,
 		GBool password=false);
-  void drawListBox(GooString **text, GBool *selection,
-		   int nOptions, int topIdx,
-		   GooString *da, GfxFontDict *fontDict, GBool quadding);
-  void layoutText(GooString *text, GooString *outBuf, int *i, GfxFont *font,
-		  double *width, double widthLimit, int *charCount,
-		  GBool noReencode);
-  void writeString(GooString *str, GooString *appearBuf);
+  void drawListBox(FormFieldChoice *fieldChoice,
+		   GooString *da, GfxResources *resources, int quadding);
 
   Form *form;
-  FormWidget *widget;                     // FormWidget object for this annotation
+  FormField *field;                       // FormField object for this annotation
   AnnotWidgetHighlightMode mode;          // H  (Default I)
   AnnotAppearanceCharacs *appearCharacs;  // MK
-  Dict *action;                           // A
-  Dict *additionActions;                  // AA
+  LinkAction *action;                     // A
+  Object additionalActions;               // AA
   // inherited  from Annot
   // AnnotBorderBS border;                // BS
   Dict *parent;                           // Parent
-  GBool regen;
   GBool addDingbatsResource;
+  Ref updatedAppearanceStream; // {-1,-1} if updateAppearanceStream has never been called
 };
 
 //------------------------------------------------------------------------
@@ -1236,19 +1397,187 @@ class Annot3D: public Annot {
   };
 public:
 
-  Annot3D(XRef *xrefA, PDFRectangle *rect, Catalog *catalog);
-  Annot3D(XRef *xrefA, Dict *dict, Catalog *catalog, Object *obj);
+  Annot3D(PDFDoc *docA, PDFRectangle *rect);
+  Annot3D(PDFDoc *docA, Dict *dict, Object *obj);
   ~Annot3D();
 
   // getters
 
 private:
 
-  void initialize(XRef *xrefA, Catalog *catalog, Dict *dict);
+  void initialize(PDFDoc *docA, Dict *dict);
 
   Activation *activation;  // 3DA
 };
 
+//------------------------------------------------------------------------
+// AnnotRichMedia
+//------------------------------------------------------------------------
+
+class AnnotRichMedia: public Annot {
+public:
+  class Params {
+  public:
+    Params(Dict *dict);
+    ~Params();
+
+    GooString* getFlashVars() const;
+
+  private:
+    // optional
+    GooString *flashVars; // FlashVars
+  };
+
+  class Instance {
+  public:
+    enum Type {
+      type3D,       // 3D
+      typeFlash,    // Flash
+      typeSound,    // Sound
+      typeVideo     // Video
+    };
+
+    Instance(Dict *dict);
+    ~Instance();
+
+    Type getType() const;
+    Params* getParams() const;
+
+  private:
+    // optional
+    Type type;     // Subtype
+    Params *params; // Params
+  };
+
+  class Configuration {
+  public:
+    enum Type {
+      type3D,       // 3D
+      typeFlash,    // Flash
+      typeSound,    // Sound
+      typeVideo     // Video
+    };
+
+    Configuration(Dict *dict);
+    ~Configuration();
+
+    Type getType() const;
+    GooString* getName() const;
+    int getInstancesCount() const;
+    Instance* getInstance(int index) const;
+
+  private:
+    // optional
+    Type type;            // Subtype
+    GooString *name;      // Name
+    Instance **instances; // Instances
+    int nInstances;
+  };
+
+  class Content;
+
+  class Asset {
+  public:
+    Asset();
+    ~Asset();
+
+    GooString* getName() const;
+    Object* getFileSpec() const;
+
+  private:
+    friend class AnnotRichMedia::Content;
+
+    GooString *name;
+    Object fileSpec;
+  };
+
+  class Content {
+  public:
+    Content(Dict *dict);
+    ~Content();
+
+    int getConfigurationsCount() const;
+    Configuration* getConfiguration(int index) const;
+
+    int getAssetsCount() const;
+    Asset* getAsset(int index) const;
+
+  private:
+    // optional
+    Configuration **configurations; // Configurations
+    int nConfigurations;
+
+    Asset **assets; // Assets
+    int nAssets;
+  };
+
+  class Activation {
+  public:
+    enum Condition {
+      conditionPageOpened,  // PO
+      conditionPageVisible, // PV
+      conditionUserAction   // XA
+    };
+
+    Activation(Dict *dict);
+
+    Condition getCondition() const;
+
+  private:
+    // optional
+    Condition condition;
+  };
+
+  class Deactivation {
+  public:
+    enum Condition {
+      conditionPageClosed,    // PC
+      conditionPageInvisible, // PI
+      conditionUserAction     // XD
+    };
+
+    Deactivation(Dict *dict);
+
+    Condition getCondition() const;
+
+  private:
+    // optional
+    Condition condition;
+  };
+
+  class Settings {
+  public:
+    Settings(Dict *dict);
+    ~Settings();
+
+    Activation* getActivation() const;
+    Deactivation* getDeactivation() const;
+
+  private:
+    // optional
+    Activation *activation;
+    Deactivation *deactivation;
+  };
+
+  AnnotRichMedia(PDFDoc *docA, PDFRectangle *rect);
+  AnnotRichMedia(PDFDoc *docA, Dict *dict, Object *obj);
+  ~AnnotRichMedia();
+
+  Content* getContent() const;
+
+  Settings* getSettings() const;
+
+private:
+  void initialize(PDFDoc *docA, Dict *dict);
+
+  // required
+  Content *content;     // RichMediaContent
+
+  // optional
+  Settings *settings;   // RichMediaSettings
+};
+
+
 //------------------------------------------------------------------------
 // Annots
 //------------------------------------------------------------------------
@@ -1256,21 +1585,25 @@ private:
 class Annots {
 public:
 
-  // Build a list of Annot objects.
-  Annots(XRef *xref, Catalog *catalog, Object *annotsObj);
+  // Build a list of Annot objects and call setPage on them
+  Annots(PDFDoc *docA, int page, Object *annotsObj);
 
   ~Annots();
 
   // Iterate through list of annotations.
   int getNumAnnots() { return nAnnots; }
   Annot *getAnnot(int i) { return annots[i]; }
+  void appendAnnot(Annot *annot);
+  GBool removeAnnot(Annot *annot);
 
 private:
-  Annot* createAnnot(XRef *xref, Dict* dict, Catalog *catalog, Object *obj);
+  Annot* createAnnot(Dict* dict, Object *obj);
   Annot *findAnnot(Ref *ref);
 
+  PDFDoc *doc;
   Annot **annots;
   int nAnnots;
+  int size;
 };
 
 #endif
